module Day12
  ( day12,
    possibleStarts,
    parseGrid,
    findPath,
    GridPoints (..),
    Grid,
    findBestPath
  )
where

import Algorithm.Search (dijkstra)
import Data.Char (ord)
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Maybe (mapMaybe)
import Data.Tuple (swap)
import Linear.V2 (V2 (..))
import Data.Sort (sortOn)

type Pos = V2 Int

type Grid = Map Pos Char

data GridPoints = GridPoints { _start :: Pos, _end :: Pos }
  deriving (Show, Eq)

parseGrid :: String -> (GridPoints, Grid)
parseGrid input =
  ( GridPoints
      { _start = inverseGrid Map.! 'S',
        _end = inverseGrid Map.! 'E'
      },
    Map.fromList grid
  )
  where
    rows = zip [0 ..] $ lines input
    grid = concatMap (uncurry colsForRow) rows
    inverseGrid = Map.fromList $ map swap grid

colsForRow :: Int -> String -> [(Pos, Char)]
colsForRow r = zipWith makePos [0 ..]
  where
    makePos c height = (V2 r c, height)

findPath :: Grid -> GridPoints -> Maybe (Int, [Pos])
findPath grid GridPoints {_start = start, _end = end} =
  dijkstra (neighbors grid) (const . const 1) solutionFound start
  where
    solutionFound = (==) end

neighbors :: Grid -> Pos -> [Pos]
neighbors grid pos = filter (canMove grid pos) possiblePositions
  where
    directions = [V2 1 0, V2 (-1) 0, V2 0 1, V2 0 (-1)]
    possiblePositions = map (+ pos) directions

canMove :: Grid -> Pos -> Pos -> Bool
canMove grid from p = case grid Map.!? p of
  Just c -> asHeight c <= (currentHeight + 1)
  Nothing -> False
  where
    asHeight 'S' = asHeight 'a'
    asHeight 'E' = asHeight 'z'
    asHeight x = ord x - ord 'a'
    currentHeight = asHeight $ grid Map.! from

possibleStarts :: Grid -> [Pos]
possibleStarts grid = map fst $ filter (\(_, v) -> v == 'a') (Map.toList grid)

findBestPath :: Grid -> GridPoints -> (Int, [Pos])
findBestPath grid points = fastest
  where
    starts = possibleStarts grid
    newPoints = [points { _start = s} | s <- starts]
    allPaths = mapMaybe (findPath grid) newPoints
    fastest = head $ sortOn fst allPaths

day12 :: IO ()
day12 = do
  input <- readFile "ressources/day12-input"
  putStrLn "Day12"
  let (positions, grid) = parseGrid input
  let path = findPath grid positions
  putStrLn ("Cost and path : " ++ show path)
  let fastest = findBestPath grid positions
  putStrLn ("Cost and path for nices route: " ++ show fastest)
