module Day7.Interpreter (
  buildTree,
  mkdir,
  FileTree (..),
  Directory,
  calculateSize,
  filterDirectories,
  sizeOfDirectories,
) where

import Day7.Parser
import Data.Map (Map)
import qualified Data.Map as Map
import qualified Data.Foldable as F


data FileTree a = FileTree
  { sub :: Map String (FileTree a)
  , files :: Map String a
  , isRoot :: Bool
} deriving (Eq)

-- Directory is a FileTree with file sizes of Int
type Directory = FileTree Int

-- pretty much only useful for summing up file sizes 
instance F.Foldable FileTree where
  foldMap f dir = foldMap f (files dir) <> mconcat mappedSubfolders
    where
      subFolders = Map.elems $ sub dir
      mappedSubfolders = map (foldMap f) subFolders

instance Show a => Show (FileTree a) where
  show d = "DIR " ++ show (Map.toList $ files d) ++ " - " ++ show (Map.toList $ sub d) ++ "\n"

mkdir :: Directory
mkdir = FileTree { sub = Map.empty, files = Map.empty, isRoot = False }

rootDirectory :: Directory
rootDirectory = mkdir { isRoot = True  }

buildTree :: [TerminalCommand] -> Directory
buildTree commands = fst (buildTree' rootDirectory commands)
  where
  -- executes terminal commands and returns all remaining commands
  buildTree' :: Directory -> [TerminalCommand] -> (Directory, [TerminalCommand])
  buildTree'  dir [] = (dir, [])
  buildTree' dir (command:cs) =  case command of
    Listing entries ->
      let asFiles = Map.fromList $ toFiles entries
          newDir = dir { files = asFiles }
      in buildTree' newDir cs -- update current directory and return all remaining commands
    In dirName ->
     let subFolders = sub dir
         selectedDir = Map.findWithDefault mkdir dirName subFolders -- select or create folder
         (replacement, rest) = buildTree' selectedDir cs -- recurse down
         updatedSub = Map.insert dirName replacement subFolders -- update folder the map of the current directory
     in buildTree' dir { sub = updatedSub} rest -- continue with updated folder and rest of commands
    Out -> (dir, cs) -- return to previous caller and return all remaining commands
    Root -> if isRoot dir
              then buildTree' dir cs -- return to previous caller as long is current directory is not root
              else (dir, command:cs) -- keep the Root command because it is not yet completed

-- convert to file-pairs and ignore directories
toFiles :: [ListingEntry] -> [(String, Int)]
toFiles ((FileListing name size):rest) = (name, size): toFiles rest
toFiles (_:rest) = toFiles rest
toFiles [] = []

calculateSize :: Directory -> Int
calculateSize = F.foldl' (+) 0

flatten :: Directory -> [Directory]
flatten d = d : concatMap flatten (Map.elems (sub d))

sizeOfDirectories :: Directory -> [Int]
sizeOfDirectories = map calculateSize . flatten

filterDirectories :: (Int -> Bool) -> Directory -> [Int]
filterDirectories predicate dir = filter predicate $ sizeOfDirectories dir

