module Day9
  ( parseMovements,
    Move (..),
    Grid (..),
    up,
    down,
    left,
    right,
    diag,
    still,
    step,
    recordPositions,
    normalizeMovement,
    Step (..),
    follow,
    buildTails,
    day9,
  )
where

import Control.Arrow ((>>>))
import Control.Monad.Trans.State.Strict (State, modify, runState)
import qualified Data.Set as S

newtype Move = Move (Int, Int)
  deriving (Show, Eq)

newtype Step = Step (Int, Int)
  deriving (Show, Eq)

up :: Int -> Move
up y = Move (0, y)

down :: Int -> Move
down y = Move (0, y * (-1))

left :: Int -> Move
left x = Move (x * (-1), 0)

right :: Int -> Move
right x = Move (x, 0)

diag :: Int -> Move
diag x = Move (x, x)

still :: Move
still = Move (0, 0)

type Pos = (Int, Int)

data Grid = Grid {h :: Pos, t :: Pos} deriving (Show, Eq)

data MovementLog = MovementLog {visited :: S.Set Pos, recordedSteps :: [Step]}
  deriving (Show, Eq)

logOne :: (Pos, Step) -> MovementLog -> MovementLog
logOne (p, s) (MovementLog v steps) =
  MovementLog (S.insert p v) (s : steps)

buildTails :: [Step] -> [([Pos], [Step])]
buildTails steps = iterate (snd >>> recordPositions) ([], steps)

recordPositions :: [Step] -> ([Pos], [Step])
recordPositions steps = (allPositions, allSteps)
  where
    -- insert last tail position
    allPositions = S.toList $ S.insert lastTailPos (visited state)
    allSteps = reverse (recordedSteps state)
    
    ((_, Grid {t = lastTailPos}), state) =
      runState (recordPositions' steps Grid {h = (0,0), t = (0,0)}) (MovementLog S.empty [])

    recordPositions' :: [Step] -> Grid -> State MovementLog ([Step], Grid)
    recordPositions' [] grid = return ([], grid)
    recordPositions' (m : ms) Grid {h = headPos, t = tailPos} =
      let newHead = headPos `step` m
          followStep = follow newHead tailPos
          newTail = tailPos `step` followStep
       in modify (logOne (newTail, followStep))
            >> recordPositions' ms Grid {h = newHead, t = newTail}

parseMovements :: String -> [Move]
parseMovements = lines >>> map toMove
  where
    toMove ('U' : ' ' : xs) = up $ read xs
    toMove ('D' : ' ' : xs) = down $ read xs
    toMove ('L' : ' ' : xs) = left $ read xs
    toMove ('R' : ' ' : xs) = right $ read xs
    toMove _ = error "cannot parse movement"

normalizeMovement :: Move -> [Step]
normalizeMovement (Move (dx, 0)) = replicate (abs dx) (Step (signum dx, 0))
normalizeMovement (Move (0, dy)) = replicate (abs dy) (Step (0, signum dy))
normalizeMovement m = error $ "cannot normalize movement: " ++ show m

step :: Pos -> Step -> Pos
step (x, y) (Step (dx, dy)) = (x + dx, y + dy)

follow :: Pos -> Pos -> Step
follow (x1, y1) (x2, y2)
  | abs dx > 1 || abs dy > 1 = Step (signum dx, signum dy)
  | otherwise = Step (0, 0)
  where
    dx = x1 - x2
    dy = y1 - y2

day9 :: IO ()
day9 = do
  input <- readFile "ressources/day09-input"
  putStrLn "Day9"
  let headSteps = parseMovements >>> concatMap normalizeMovement $ input
  let allTails = buildTails headSteps
  let positions = (!! 1) >>> fst >>> length $ allTails
  putStrLn ("Number of distinct positions " ++ show positions)
  let tail9 = (!! 9) >>> fst >>> length $ allTails
  putStrLn ("Number of distinct positions of tail 9" ++ show tail9)
